from __future__ import annotations

import asyncio
from typing import Any, Callable, TypeVar, overload
from collections.abc import Sequence, Mapping
from weakref import finalize
from itertools import chain

from .event import BaseEvent
from .auxiliary import BaseAuxiliary
from .context import publisher_ctx
from .handler import dispatch
from .provider import Provider, ProviderFactory
from .publisher import BackendPublisher, ExternalPublisher, Publisher
from .subscriber import Subscriber
from .typing import Result, Resultable

T = TypeVar("T")


class EventSystem:
    _ref_tasks = set()
    _backend_publisher: Publisher = BackendPublisher("__backend__publisher__")
    publishers: dict[str, Publisher]

    def __init__(self):
        self.publishers = {}

        def _remove(es):
            for task in es._ref_tasks:
                if not task.done():
                    task.cancel()
            es._ref_tasks.clear()

        finalize(self, _remove, self)

    async def setup_fetch(self):
        self._ref_tasks.add(asyncio.create_task(self._loop_fetch()))

    async def _loop_fetch(self):
        while True:
            for publisher in self.publishers.values():
                if not (event := (await publisher.supply())):
                    continue
                self.post(event, publisher)
            await asyncio.sleep(0.05)

    def register(self, *publishers: Publisher):
        """注册发布者"""
        for publisher in publishers:
            self.publishers[publisher.id] = publisher

    def define(
        self,
        name: str,
        target: type[T] | None = None,
        supplier: Callable[[T], Mapping[str, Any]] | None = None,
        predicate: Callable[[T], bool] | None = None,
    ) -> Publisher:
        if name in self.publishers:
            return self.publishers[name]
        elif predicate and (_key := f"{name}::{predicate}") in self.publishers:
            return self.publishers[_key]
        if not target:
            publisher = BackendPublisher(name, predicate)
        elif issubclass(target, BaseEvent):
            publisher = Publisher(name, target, predicate=predicate)
        else:
            publisher = ExternalPublisher(name, target, supplier, predicate)
        self.register(publisher)
        return publisher

    def publish(self, event: Any, publisher: str | Publisher | None = None):
        """发布事件"""
        loop = asyncio.get_running_loop()
        if isinstance(publisher, str) and (pub := self.publishers.get(publisher)):
            task = loop.create_task(dispatch(pub.subscribers.values(), event))
            self._ref_tasks.add(task)
            task.add_done_callback(self._ref_tasks.discard)
            return task
        if isinstance(publisher, Publisher):
            task = loop.create_task(dispatch(publisher.subscribers.values(), event))
            self._ref_tasks.add(task)
            task.add_done_callback(self._ref_tasks.discard)
            return task
        task = loop.create_task(
            dispatch(
                chain.from_iterable(
                    pub.subscribers.values() for pub in self.publishers.values() if pub.validate(event)
                ),
                event
            )
        )
        self._ref_tasks.add(task)
        task.add_done_callback(self._ref_tasks.discard)
        return task

    @overload
    def post(self, event: Resultable[T], publisher: str | Publisher | None = None) -> asyncio.Task[Result[T] | None]:
        ...

    @overload
    def post(self, event: Any, publisher: str | Publisher | None = None) -> asyncio.Task[Result[Any] | None]:
        ...

    def post(self, event: Any, publisher: str | Publisher | None = None):
        """发布事件并返回第一个响应结果"""
        loop = asyncio.get_running_loop()
        if isinstance(publisher, str) and (pub := self.publishers.get(publisher)):
            task = loop.create_task(dispatch(pub.subscribers.values(), event, return_result=True))
            self._ref_tasks.add(task)
            task.add_done_callback(self._ref_tasks.discard)
            return task
        if isinstance(publisher, Publisher):
            task = loop.create_task(dispatch(publisher.subscribers.values(), event, return_result=True))
            self._ref_tasks.add(task)
            task.add_done_callback(self._ref_tasks.discard)
            return task
        task = loop.create_task(
            dispatch(
                chain.from_iterable(
                    pub.subscribers.values() for pub in self.publishers.values() if pub.validate(event)
                ),
                event,
                return_result=True
            )
        )
        self._ref_tasks.add(task)
        task.add_done_callback(self._ref_tasks.discard)
        return task

    @overload
    def on(
        self,
        events: type | tuple[type, ...],
        func: Callable[..., Any],
        *,
        priority: int = 16,
        auxiliaries: list[BaseAuxiliary] | None = None,
        providers: Sequence[
            Provider[Any] | type[Provider[Any]] | ProviderFactory | type[ProviderFactory]
        ] | None = None,
    ) -> Subscriber:
        ...

    @overload
    def on(
        self,
        events: type | tuple[type, ...],
        *,
        priority: int = 16,
        auxiliaries: list[BaseAuxiliary] | None = None,
        providers: Sequence[
            Provider[Any] | type[Provider[Any]] | ProviderFactory | type[ProviderFactory]
        ] | None = None,
    ) -> Callable[[Callable[..., Any]], Subscriber]:
        ...

    @overload
    def on(
        self,
        *,
        priority: int = 16,
        auxiliaries: list[BaseAuxiliary] | None = None,
        providers: Sequence[
            Provider[Any] | type[Provider[Any]] | ProviderFactory | type[ProviderFactory]
        ] | None = None,
    ) -> Callable[[Callable[..., Any]], Subscriber]:
        ...

    def on(
        self,
        events: type | tuple[type, ...] | None = None,
        func: Callable[..., Any] | None = None,
        priority: int = 16,
        auxiliaries: list[BaseAuxiliary] | None = None,
        providers: Sequence[
            Provider[Any] | type[Provider[Any]] | ProviderFactory | type[ProviderFactory]
        ] | None = None,
    ):
        if not (pub := publisher_ctx.get()):
            if not events:
                pub = self._backend_publisher
            else:
                events = events if isinstance(events, tuple) else (events,)
                pub = Publisher(f"global::{sorted(events, key=lambda e: id(e))}", *events)
        if pub.id in self.publishers:
            pub = self.publishers[pub.id]
        else:
            self.publishers[pub.id] = pub

        if not func:
            return pub.register(priority=priority, auxiliaries=auxiliaries, providers=providers)
        return pub.register(func, priority=priority, auxiliaries=auxiliaries, providers=providers)

    @overload
    def use(
        self,
        pub: str | Publisher,
        func: Callable[..., Any],
        *,
        priority: int = 16,
        auxiliaries: list[BaseAuxiliary] | None = None,
        providers: Sequence[
           Provider[Any] | type[Provider[Any]] | ProviderFactory | type[ProviderFactory]
        ] | None = None,
    ) -> Subscriber:
        ...

    @overload
    def use(
        self,
        pub: str | Publisher,
        *,
        priority: int = 16,
        auxiliaries: list[BaseAuxiliary] | None = None,
        providers: Sequence[
           Provider[Any] | type[Provider[Any]] | ProviderFactory | type[ProviderFactory]
        ] | None = None,
    ) -> Callable[[Callable[..., Any]], Subscriber]:
        ...

    @overload
    def use(
        self,
        *,
        priority: int = 16,
        auxiliaries: list[BaseAuxiliary] | None = None,
        providers: Sequence[
           Provider[Any] | type[Provider[Any]] | ProviderFactory | type[ProviderFactory]
        ] | None = None,
    ) -> Callable[[Callable[..., Any]], Subscriber]:
        ...

    def use(
        self,
        pub: str | Publisher | None = None,
        func: Callable[..., Any] | None = None,
        priority: int = 16,
        auxiliaries: list[BaseAuxiliary] | None = None,
        providers: Sequence[
           Provider[Any] | type[Provider[Any]] | ProviderFactory | type[ProviderFactory]
        ] | None = None,
    ):
        if not pub:
            publisher = publisher_ctx.get() or self._backend_publisher
        else:
            publisher = pub if isinstance(pub, Publisher) else self.publishers[pub]
        if not func:
            return publisher.register(priority=priority, auxiliaries=auxiliaries, providers=providers)
        return publisher.register(func, priority=priority, auxiliaries=auxiliaries, providers=providers)


es = EventSystem()
